# SPDX-FileCopyrightText: 2021 Daniel Vrátil <dvratil@kde.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
# SPDX-License-Identifier: LGPL-3.0-only
# SPDX-License-Identifier: LicenseRef-KDE-Accepted-LGPL

from typing import Dict, List
from functools import partial

from schema_types import Object, Ref, Type


class ParserError(Exception):
    """
    An exception that is raised when the schema cannot be parsed.
    """


class RefVisitor:
    """
    RefVisitor is a helper class for binding Refs with referenced types.

    A Ref type is a reference to another type (usually Object) defined elsewhere
    in the schema. Each Ref type is registered into the RefVisitor. Once the entire
    document is parsed, the RefVisitor is called to update all Refs with links to
    definitions of the actual type.
    """

    def __init__(self, api):
        self._refs: List[Ref] = []
        self._api = api

    def add_ref(self, ref: "Ref") -> None:
        self._refs.append(ref)

    def _update_ref_types(self, schemas: List[Type]) -> None:
        def match_ref(ref: Ref, schema: Object):
            return ref.name == schema.name

        for ref in self._refs:
            # pylint: disable=protected-access # friend access
            ref._ref_type = next(filter(partial(match_ref, ref), schemas), None)  # type: ignore
            if not ref._ref_type:
                raise RuntimeError(f"Failed to find type for ref {ref.name}.")


    def _handle_renamed_objects(self, schemas: List[Type]) -> None:
        for old_name, new_name in self._api["renameObjects"].items():
            for ref in self._refs:
                if ref.name == old_name:
                    ref.rename_to(new_name)

    def update_refs(self, schemas: List[Type]) -> None:
        self._handle_renamed_objects(schemas)
        self._update_ref_types(schemas)


class Parser:
    """
    Parser for the schema described in the Discovery document.
    """

    def __init__(self, api: dict, schema: dict):
        self._api = api
        self._schema = schema

    def _handle_renamed_objects(self, obj: Object):
        for old_name, new_name in self._api["renameObjects"].items():
            if obj.name == old_name:
                obj.rename_to(new_name)

    def _parse_schema_type(self, schema_name: str, schema: Dict, ref_visitor):
        if schema["type"] == "object":
            obj = Object(schema, ref_visitor)
            self._handle_renamed_objects(obj)
            return obj

        raise ParserError(
            f"Unknown schema type \"{schema['type']}\" for schema {schema_name}"
        )

    def parse_schema(self) -> List[Object]:
        ref_visitor = RefVisitor(self._api)
        schemas = list(
            map(
                lambda item: self._parse_schema_type(
                    item[0], item[1], ref_visitor=ref_visitor
                ),
                self._schema["schemas"].items(),
            )
        )
        ref_visitor.update_refs(schemas)

        return schemas



#!python3
from typing import List
import requests
import textwrap
import itertools

#from property import Property
from schema_types import *

def generate_schema_code(schema, cppfile, hppfile, kobject=True):
    properties = list(map(lambda p: Property(*p), schema['properties'].items()))
    with open(hppfile, 'w') as header, open(cppfile, 'w') as source:
        header.write(textwrap.dedent("""\
            #pragma once

            // This is a generated file. Do not edit!
        """))
        if kobject:
            header.write("#include \"object.h\"\n")
        else:
            header.write('#include <QSharedDataPointer>\n')

        includes = set(itertools.chain.from_iterable(map(lambda p: p.includes, properties)))
        for include in includes:
            header.write(f"#include {include}\n")

        header.write("\nnamespace KGAPI2::People {\n\n")
        classname = schema['id'][0].upper() + schema['id'][1:]
        if kobject:
            header.write(textwrap.dedent(f"""\
                class {classname} : public KGAPI2::Object {{
                    Q_OBJECT
            """))
        else:
            header.write(textwrap.dedent(f"""\
                class {classname} {{
                    Q_GADGET
            """))

        for prop in properties:
            header.write(f"    /** {prop.description} **/\n")
            header.write(f"    {prop.qproperty_string(notify=kobject)}\n")

        # First, write nested types
        header.write(textwrap.dedent(f"""\
            public:
                explicit {schema['id']}();
                ~{schema['id']}();
        """))

        for prop in properties:
            header.write("\n")
            header.write(f"    {prop.type} {prop.getter}() const;\n")
            if not prop.read_only:
                header.write(f"    void {prop.setter}(const {prop.type} &{prop.name});\n")
            if prop.is_array:
                var_name = f"{prop.array_type[0].lower()}{prop.array_type[1:]}"
                header.write(f"    void add{prop.array_type}(const {prop.array_type} &{var_name});\n")
                header.write(f"    void remove{prop.array_type}(const {prop.array_type} &{var_name});\n")
                header.write(f"    void clear{prop.name[0].upper()}{prop.name[1:]}();\n")

        ###### Privaet part
        header.write(textwrap.dedent(f"""\
            private:
                class Private;
        """))

        if kobject:
            header.write('    Private const *d;\n')
        else:
            header.write('    const QSharedDataPointer<Private> d;\n')

        header.write(textwrap.dedent(f"""\
            }}; // class {schema['id']}

            }} // namespace KGAPI2::People
        """))



def parse_enums(schemas: dict):
    def find_enum(subtree):
        for name, value in subtree.items():
            if name == 'properties':
                for type, type_def in value.items():
                    if 'enum' in type_def:
                        values = type_def['enum']
                        # v contains enum, p is the name of the prop
                # look for enum props now
            elif isinstance(value, dict):
                find_enum(value)


def _parse_object_schema(schema):
    o = Object()
    
def _parse_schema_type(schema):
    if schema['type'] == 'object':
        return Object(schema)
    else:
        print(f"Unknown schema type {schema['type']}")
        

def parse_schema(uri):
    disc_req = requests.get(uri)
    disc_doc = disc_req.json()

    results = []
    for name, schema in disc_doc['schemas'].items():
        # Skip schemas representing API call requests/respones
        if name.endswith('Request') or name.endswith('Response'):
            continue

        results.append(_parse_schema_type(schema))

    return results


import argparse
from parser import parse_schema

def main():
    argparser = argparse.ArgumentParser(add_help=True)
    argparser.add_argument('--discovery', metavar='FILE_OR_URL', help='File or URL of the Discovery Document', required=True)
    argparser.add_argument('--outdir', metavar='DIR', help='Output directory', default='.')
    args = argparser.parse_args()


    schema = parse_schema(args.discovery)

    print(schema)

if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2021 Daniel Vrátil <dvratil@kde.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
# SPDX-License-Identifier: LGPL-3.0-only
# SPDX-License-Identifier: LicenseRef-KDE-Accepted-LGPL